#include "ruby.h"

static VALUE cWriteBuffer;

struct WriteBuffer {
  unsigned char *ptr;
  unsigned long size;
  unsigned long capacity;
  unsigned long init_size;
  unsigned long next_size;
};

static void wbuf_free(struct WriteBuffer *obj) {
  void *ptr;
  
  ruby_xfree(obj->ptr);
  ruby_xfree(obj);
  
}

static VALUE wbuf_new(VALUE self, VALUE v_init_size, VALUE v_next_size) {
  struct WriteBuffer *ptr;
  unsigned long init_size;
  unsigned long next_size;
  VALUE volatile obj;
  
  Check_Type(v_init_size, T_FIXNUM);
  Check_Type(v_next_size, T_FIXNUM);
  
  init_size = FIX2INT(v_init_size);
  next_size = FIX2INT(v_next_size);
  
  obj = Data_Make_Struct(cWriteBuffer, struct WriteBuffer, NULL, wbuf_free, ptr);
  
  ptr->ptr       = ALLOC_N(char, init_size);
  ptr->size      = 0;
  ptr->capacity  = init_size;
  ptr->init_size = init_size;
  ptr->next_size = next_size;
  
  return obj;
}

static VALUE wbuf_size(VALUE self) {
  struct WriteBuffer *wbuf;
  
  Data_Get_Struct(self, struct WriteBuffer, wbuf);
  
  return INT2FIX(wbuf->size);
}

static VALUE wbuf_capacity(VALUE self) {
  struct WriteBuffer *wbuf;
  
  Data_Get_Struct(self, struct WriteBuffer, wbuf);
  
  return INT2FIX(wbuf->capacity);
}

static VALUE wbuf_clear(VALUE self) {
  struct WriteBuffer *wbuf;

  Data_Get_Struct(self, struct WriteBuffer, wbuf);
  wbuf->ptr       = REALLOC_N(wbuf->ptr, char, wbuf->init_size);
  wbuf->size      = 0;
  wbuf->capacity  = wbuf->init_size;
  
  return self;
}


static VALUE wbuf_concat(VALUE self, VALUE str) {
  struct WriteBuffer *wbuf;
  unsigned char *src, *dst;
  unsigned long str_size;
  unsigned long new_size;
  unsigned long new_capacity;
  
  Check_Type(str, T_STRING);
  OBJ_INFECT(self, str);
  Data_Get_Struct(self, struct WriteBuffer, wbuf);
  
  str_size = RSTRING(str)->len;
  if (str_size==0) {
    return self;
  }
  
  new_size = wbuf->size + str_size;
  if ( new_size > wbuf->capacity ) {
    wbuf->capacity = ((new_size + wbuf->next_size) / wbuf->next_size) * wbuf->next_size;
    REALLOC_N( wbuf->ptr , char, wbuf->capacity);
  }
  src = RSTRING(str)->ptr;
  dst = wbuf->ptr + wbuf->size;
  if (str_size<4) {
    while (str_size--) {
      *dst++ = *src++;
    }
  } else {
    memcpy(dst , src , str_size);
  }
  wbuf->size = new_size;
  
  return self;
}

static VALUE wbuf_string(VALUE self) {
  struct WriteBuffer *wbuf;
  
  Data_Get_Struct(self, struct WriteBuffer, wbuf);
  
  return rb_str_new(wbuf->ptr, wbuf->size);
}


void Init_writebuffer() {
  ID id;
  VALUE mWeb;

  id = rb_intern("Web");
  if (rb_const_defined(rb_cObject, id)) {
    mWeb = rb_const_get(rb_cObject, id);
  } else {
    mWeb = rb_define_module("Web");
  }
  cWriteBuffer =  rb_define_class_under(mWeb, "WriteBuffer", rb_cObject);
  
  rb_define_singleton_method( cWriteBuffer, "new", wbuf_new, 2 );
  rb_define_method( cWriteBuffer, "size", wbuf_size, 0 );
  rb_define_method( cWriteBuffer, "capacity", wbuf_capacity, 0 );
  rb_define_method( cWriteBuffer, "<<", wbuf_concat, 1 );
  rb_define_method( cWriteBuffer, "write", wbuf_concat, 1 );
  rb_define_method( cWriteBuffer, "clear", wbuf_clear, 0 );
  rb_define_method( cWriteBuffer, "to_s", wbuf_string, 0 );
  
}
